#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital
import numpy as np
import pmt

from ... import viterbi_decoder
from ...crcs import crc16_ccitt_x25
from ...hier.sync_to_pdu import sync_to_pdu
from ...utils.options_block import options_block


# Syncword is 0xe37b167e followed by a 1
# (UART stop bit for 0x7e)
# This is UART encoded
_syncword = '111000110111101100010110011111101'


_scrambler_seq = np.array(
    [0x0e, 0xf2, 0xc9, 0x02, 0x26, 0x2e, 0xb6, 0x0c, 0xd4, 0xe7,
     0xb4, 0x2a, 0xfa, 0x51, 0xb8, 0xfe, 0x1d, 0xe5, 0x92, 0x04,
     0x4c, 0x5d, 0x6c, 0x19, 0xa9, 0xcf, 0x68, 0x55, 0xf4, 0xa3,
     0x71, 0xfc, 0x3b, 0xcb, 0x24, 0x08, 0x98, 0xba, 0xd8, 0x33,
     0x53, 0x9e, 0xd0, 0xab, 0xe9, 0x46, 0xe3, 0xf8, 0x77, 0x96,
     0x48, 0x11, 0x31, 0x75, 0xb0, 0x66, 0xa7, 0x3d, 0xa1, 0x57,
     0xd2, 0x8d, 0xc7, 0xf0, 0xef, 0x2c, 0x90, 0x22, 0x62, 0xeb,
     0x60, 0xcd, 0x4e, 0x7b, 0x42, 0xaf, 0xa5, 0x1b, 0x8f, 0xe1,
     0xde, 0x59, 0x20, 0x44, 0xc5, 0xd6, 0xc1, 0x9a, 0x9c, 0xf6,
     0x85, 0x5f, 0x4a, 0x37, 0x1f, 0xc3, 0xbc, 0xb2, 0x40, 0x89,
     0x8b, 0xad, 0x83, 0x35, 0x39, 0xed, 0x0a, 0xbe, 0x94, 0x6e,
     0x3f, 0x87, 0x79, 0x64, 0x81, 0x13, 0x17, 0x5b, 0x06, 0x6a,
     0x73, 0xda, 0x15, 0x7d, 0x28, 0xdc, 0x7f],
    dtype='uint8')


class prepare_fec(gr.basic_block):
    """
    Helper block to prepare the convolutional encoded FEC codeword.

    This performs UART 8N1 decoding, deinterleaving and descrambling.
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='prepare_fec',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = np.array(pmt.u8vector_elements(msg), dtype='uint8')
        assert packet.size % 10 == 0

        uart_decode = np.packbits(
            packet.reshape(-1, 10)[:, 1:-1][:, ::-1])
        fec_data = uart_decode[16:16+102]
        # Deinterleave: the last 6 bytes are not interleaved, since the
        # interleaver works in blocks of 8 bytes
        data_a, data_b = fec_data[:-6], fec_data[-6:]
        data_a = np.packbits(
            np.einsum('ijk->ikj',
                      np.unpackbits(data_a).reshape(-1, 8, 8)[:, :, ::-1])
            [:, :, ::-1])
        deinter = np.concatenate((data_a, data_b))
        descrambled = deinter ^ _scrambler_seq[:102]

        # The input to the convolutional code always ends with a 0x00 byte to
        # terminate the trellis. This produces 16 encoded bits. However, only
        # 2 bits are needed to terminate the trellis (K = 3), which produce 4
        # encoded bits. Thus, we can throw away the last 12 bits.
        viterbi_in = np.unpackbits(descrambled)[:-12]

        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.car(msg_pmt),
                     pmt.init_u8vector(len(viterbi_in), viterbi_in)))


class pack_viterbi_out(gr.basic_block):
    """
    Helper block to pack the Viterbi decoder output.
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='pack_viterbi_out',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = np.array(pmt.u8vector_elements(msg), dtype='uint8')
        packet = np.packbits(packet)
        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.car(msg_pmt),
                     pmt.init_u8vector(len(packet), packet)))


class hsu_sat1_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe HSU-SAT1 frames.

    HSU-SAT1 uses custom frames with an 8N1 UART encoding and FEC
    following the library https://github.com/f4goh/CONVOLUTION
    This consists of an r=1/2 k=3 convolutional code, block interleaver
    and scrambler.

    The input is a float stream of soft symbols. The output are PDUs
    with HSU-SAT1 frames. Only the FEC data (with the CRC and Viterbi
    tail stripped) is included in the output.

    Args:
        options: Options from argparse
    """
    def __init__(self, options=None):
        gr.hier_block2.__init__(
            self,
            'hsu_sat1_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        self.slicer = digital.binary_slicer_fb()
        self.deframer = sync_to_pdu(
            packlen=(16+102)*10, sync=_syncword,
            threshold=self.options.syncword_threshold)
        self.prepare = prepare_fec()
        self.viterbi = viterbi_decoder(3, [7, 5])
        self.pack = pack_viterbi_out()
        self.crc = crc16_ccitt_x25(swap_endianness=False)

        self.connect(self, self.slicer, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.prepare, 'in'))
        self.msg_connect((self.prepare, 'out'), (self.viterbi, 'in'))
        self.msg_connect((self.viterbi, 'out'), (self.pack, 'in'))
        self.msg_connect((self.pack, 'out'), (self.crc, 'in'))
        self.msg_connect((self.crc, 'ok'), (self, 'out'))

    _default_sync_threshold = 4

    @classmethod
    def add_options(cls, parser):
        """
        Adds DIY-1 deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt

# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites import hdlc_framer, hdlc_deframer
from satellites.grtypes import byte_t


class qa_hdlc(gr_unittest.TestCase):
    def setUp(self):
        self.tb = gr.top_block()

    def tearDown(self):
        self.tb = None

    def test_framer_deframer(self):
        """Connects an HDLC framer to a deframer and sends PDUs through"""
        framer = hdlc_framer(100, 20)
        deframer = hdlc_deframer(True, 10000)
        pdu2tag = blocks.pdu_to_tagged_stream(byte_t)
        dbg = blocks.message_debug()

        self.tb.connect(pdu2tag, deframer)
        self.tb.msg_connect((framer, 'out'), (pdu2tag, 'pdus'))
        self.tb.msg_connect((deframer, 'out'), (dbg, 'store'))

        test_size = 150
        test_number_frames = 7
        test_data = [bytes(np.random.randint(0, 256, test_size, dtype='uint8'))
                     for _ in range(test_number_frames)]
        for td in test_data:
            test_frame = pmt.cons(pmt.PMT_NIL,
                                  pmt.init_u8vector(test_size, list(td)))
            framer.to_basic_block()._post(pmt.intern('in'), test_frame)
        framer.to_basic_block()._post(
            pmt.intern('system'),
            pmt.cons(pmt.intern('done'), pmt.from_long(1)))

        self.tb.start()
        self.tb.wait()

        for j, td in enumerate(test_data):
            result_data = bytes(
                pmt.u8vector_elements(pmt.cdr(dbg.get_message(j))))
            self.assertEqual(
                td, result_data,
                'HDLC deframer output does not match expected frame')


if __name__ == '__main__':
    gr_unittest.run(qa_hdlc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2023 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct

from .imagereceiver import ImageReceiver

# Documentation:
# https://ukamsat.files.wordpress.com/2022/12/
# camsat-cas-5a-amateur-radio-satellite-users-manual-v2.0.pdf


class ImageReceiverCAS5A(ImageReceiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_fid = -1
        self._first_chunk = None
        self._current_offset = 0
        self._current_chunk = None
        self._next_offset = 0

    def parse_chunk(self, chunk):
        # The AX.25 header occupies 16 bytes
        if len(chunk) <= 16:
            return None
        chunk = chunk[16:]
        if chunk.startswith(b'\x01\x00' * 3):
            # Telemetry frames start by 01 00 01 00 01 00.
            # The documentation says that they start by
            # 01 00 01 00 01 00 7E,
            # but they actually seem  to start by
            # 01 00 01 00 01 00 A7
            return None
        return chunk

    def chunk_data(self, chunk):
        return chunk

    def chunk_offset(self, chunk):
        # To determine the offset, we count how many bytes we've seen so far
        # (this assumes we don't lose any chunks).
        if self._current_chunk != chunk:
            self._current_chunk = chunk
            self._current_offset = self._next_offset
        offset = self._current_offset
        self._next_offset = offset + len(chunk)
        return offset

    def file_id(self, chunk):
        # A new JPEG file starts by ff d8
        if chunk.startswith(b'\xff\xd8') and chunk != self._first_chunk:
            self._current_fid += 1
            self._current_offset = 0
            self._first_chunk = chunk
            self._current_chunk = chunk
            self._next_offset = 0
        return self._current_fid


class ImageReceiverCAS5ANew(ImageReceiver):
    # Documentation:
    # https://twitter.com/scott23192/status/1630749478561935360
    # https://mega.nz/file/4rIywT5L#WoZsMxzIkUKhqHTrYh__nvv_N9CGwVV-dLsJ2k4_2OA

    def parse_chunk(self, chunk):
        # The AX.25 header occupies 16 bytes
        if len(chunk) <= 16:
            return None
        chunk = chunk[16:]
        if not chunk.startswith(b'\x03'):
            return None
        return chunk

    def chunk_sequence(self, chunk):
        # sequences are 1-based rather than 0-based
        return struct.unpack('>H', chunk[3:5])[0] - 1

    def chunk_data(self, chunk):
        return chunk[16:]

    def chunk_size(self):
        # This isn't stated explicitly in the documentation, but it seems that
        # all the chunks except the last one are 240 bytes.
        return 240

    def is_last_chunk(self, chunk):
        num_chunks = struct.unpack('>H', chunk[1:3])[0]
        return num_chunks == self.chunk_sequence(chunk) + 1


cas5a = ImageReceiverCAS5A
cas5a_new = ImageReceiverCAS5ANew



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt

# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites import pdu_head_tail


class qa_pdu_head_tail(gr_unittest.TestCase):
    def test_pdu_head_tail(self):
        for mode in range(4):
            with self.subTest(mode=mode):
                self.run_mode(mode)

    def run_mode(self, mode):
        tb = gr.top_block()
        dbg = blocks.message_debug()
        num = 10
        sizes = [15, 8]
        test_data = list(range(np.max(sizes)))
        msgs = [test_data[:x] for x in sizes]
        pdus = [pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(m), m))
                for m in msgs]
        head_tail = pdu_head_tail(mode, num)

        tb.msg_connect((head_tail, 'out'), (dbg, 'store'))
        for pdu in pdus:
            head_tail.to_basic_block()._post(pmt.intern('in'), pdu)
        head_tail.to_basic_block()._post(
            pmt.intern('system'),
            pmt.cons(pmt.intern('done'), pmt.from_long(1)))

        tb.start()
        tb.wait()

        for j, msg in enumerate(msgs):
            out = pmt.u8vector_elements(pmt.cdr(dbg.get_message(j)))
            if mode == 0:
                expected = msg[:num]
            elif mode == 1:
                expected = msg[:-num]
            elif mode == 2:
                expected = msg[-num:]
            elif mode == 3:
                expected = msg[num:]
            else:
                raise ValueError('Invalid mode')
            self.assertEqual(
                out, expected,
                "PDU head/tail output does not match expected value")


if __name__ == '__main__':
    gr_unittest.run(qa_pdu_head_tail)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt

# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites import pdu_length_filter


class qa_pdu_length_filter(gr_unittest.TestCase):
    def test_pdu_length_filter(self):
        tb = gr.top_block()
        dbg = blocks.message_debug()
        sizes = [25, 50, 100]
        test_data = list(range(np.max(sizes)))
        msgs = [test_data[:x] for x in sizes]
        pdus = [pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(m), m))
                for m in msgs]
        length_filter = pdu_length_filter(40, 60)

        tb.msg_connect((length_filter, 'out'), (dbg, 'store'))
        for pdu in pdus:
            length_filter.to_basic_block()._post(pmt.intern('in'), pdu)
        length_filter.to_basic_block()._post(
            pmt.intern('system'),
            pmt.cons(pmt.intern('done'), pmt.from_long(1)))

        tb.start()
        tb.wait()

        self.assertEqual(
            dbg.num_messages(), 1,
            'Incorrect number of messages passed by PDU Length Filter')
        out = pmt.u8vector_elements(pmt.cdr(dbg.get_message(0)))
        self.assertEqual(
            len(out), 50,
            'PDU Length Filter output does not match expected')


if __name__ == '__main__':
    gr_unittest.run(qa_pdu_length_filter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt

# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites.components.deframers import usp_deframer


class qa_usp_deframer(gr_unittest.TestCase):
    def setUp(self):
        frame_long = """
24 c8 d6 9c 06 17 78 af 8c f5 8d 82 57 a5 36 8e
58 c2 fa 4f ec bd 1a 64 95 5b 91 77 97 ce a8 09
f5 26 3e 74 a1 21 96 4c 8a 08 11 ec cf 08 8f bc
30 9f 88 7f 77 d3 8f b2 cf 77 07 59 c7 09 47 16
a5 ce 36 59 e8 ee 9a c7 80 69 23 35 26 3e ad 3a
e4 53 21 08 12 bc 1f 43 46 4d c6 c2 8b e2 27 7c
4f fb f6 09 50 5e de 51 1c 3c 90 0e 66 e1 58 aa
0c 5b d5 da 18 75 f9 cb 3d d2 ec a7 f3 02 db 5a
97 38 d9 67 a3 ba 6b 1e 01 a4 8c d4 98 fa b4 eb
91 4c 84 20 4a f0 7d 0d 19 37 1b 0a 2c d6 bf 1f
c7 10 cb 95 08 ae 91 05 be 3b fb 93 50 49 ad 3a
06 d8 62 5e 78 7a 9d b5 9c 48 0e ef c2 c1 6a 6a
85 75 d6 f6 f2 54 ce c0 c9 8a 95 cb e4 4a a3 74
5a f1 3c 68 6f 25 79 38 49 c8 8f b1 9f 92 77 c4
ff bf 60 95 05 ed e5 11 c3 c9 00 e6 6e 15 8a a0
c5 bd 5d a1 87 5f 9c b3 dd 2e ca 7f 30 2d b5 a9
73 8d 96 7a 3b a6 b1 e0 1a 48 cd 49 8f ab 4e b9
14 c8 42 04 af 07 d0 d1 93 71 b0 a2 f8 89 df 13
fe fd 82 54 17 b7 94 47 0f 24 03 99 b8 56 2a 83
16 f5 76 86 1d 7e 72 cf 74 bb 29 fc c0 b6 d6 a5
ce 36 59 e8 ee 9a c7 80 69 23 35 26 3e ad 3a e4
53 21 08 12 bc 1f 43 46 4d c6 c2 8b e2 27 7c 4f
fb f6 09 50 5e de 51 1c 3c 90 0e 66 e1 58 aa 0c
5b d5 da 18 75 f9 cb 3d d2 ec a7 f3 02 db 5a 97
38 d9 67 a3 ba 6b 1e 01 a4 8c d4 98 fa b4 eb 91
4c 84 20 4a f0 7d 0d 19 37 1b 0a 2f 88 9d f1 3f
ef d8 25 41 7b 79 44 70 f2 40 39 9b 85 62 a8 31
6f 57 68 61 d7 e7 2c f7 4b b2 9f cc 0b 6d 6a 5c
e3 65 9e 8e e9 ac 75 48 a5 8a 95 f3 a4 12 87 01
0f 08 58 d8 fc 00 a1 9e 32 26 33 e3 00 c4 10 a0
df b5 3a b4 07 09 49 0b 27 d9 37 cf e9 59 8d 8c
93 65 49 05 01 70 b8 c5 d8 07 d0 ea b5 a4 06 46
eb 3a c8 05 f9 f3
"""
        frame_long = frame_long.replace(' ', '').replace('\n', '')
        self.frame_long = np.frombuffer(bytes.fromhex(frame_long),
                                        dtype='uint8')

        frame_short = """
71 9d 83 c9 53 42 2d fa 8c f5 8d 82 6c 61 8a fe
58 c2 fa 4f ec bd 1a 64 95 5b 91 77 97 ce a8 09
f5 26 3e 74 a1 21 96 4c 8a 08 11 ec cf 08 8f bc
e9 3f 50 49 a0 a3 8f b2 cf 77 07 59 fc cd fb 66
a5 ce 38 9d 1e 1b a1 9d 6f 57 8a 30 b3 04 b8 42
0a 8a c5 48 7f 98 18 3f a8 1d d8 8d 4f da 01 cc
4f fb f6 09 50 5e de 51 cb ba 35 b5 15 53 35 10
56 42 4d 4c bc b4 1e 52 4f 8a 5b f4 f7 2e 07 c9
5d 1e 20 5e d0 4c 7f 1f 13 1c 76 15 d9 46 b6 2d
75 51 c5 7e f8 50 72 67 42 5a 04 1e 6a e0 1a 7e
9d d0 29 f1 57 2c 20 fb 9e 20 1e fe ff ff f9 12
a0 d6 2d 46 76 1f 8d ee de f8 63 67 20 8c 8c bb
05 a7 95 0a d9 8b 4c 14 19 7b 07 29 46 99 5e 3c
e5 e2 da ab 03 7c b5 25 53 14 ad e7 2b 61 ff 37
9e c8 d1 b3 8e db f8 6d 2c db f7 8e ed 31 b6 ab
5f 34 27 6e a5 1f 7b 31 b5 eb 6e 74 23 dc 5a bf
45 41 53 64 ad d5 9d 59 f2 2e e5 39 d6 e8 de b5
6b da 8b 84 3d b1 56 9f 51 4a c9 9d c6 34 7a aa
07 e4 bc 43 43 3d c5 9b 5b f1 8b 58 54 53 bb 59
c6 72 57 31 e3 83 af 4d d4 c5 68 4e 6e 7a 15 36
9e 3a 46 d0 4b c8 ae 7d 1b c4 f4 eb 65 87 49 ba
6f 9a 51 90 22 9e 12 95 ec b4 42 4a 85 76 5e 4f
49 50 83 4f 26 ea 29 5d a9 56 d3 24 b2 5c cd 1b
ba 9e db d2 3b 72 1d 16 5b d1 57 72 ec 21 e6 7d
91 57 56 8a 0a dd 9a c2 a2 b7 b5 3d bb 11 bc 82
e9 9a 83 8d 8f 90 14 1b da c4 56 ca 2b b3 aa 9a
7b d2 49 2e cd 46 78 aa 6b f0 66 0f 64 cb 16 60
07 e6 55 12 67 2d 18 6a 17 12 92 db 25 97 69 45
b5 43 16 93 56 94 75 a1 fa a2 67 4e e8 d8 42 87
3b 2b e0 6b 5b 17 06 17 54 e4 c4 d6 d6 4b 7d 22
4f 33 2f b2 d2 59 45 49 ca b4 a5 53 c2 96 21 29
ec 88 a9 37 c3 0b 5e 58 c5 26 9c c2 ae a4 c4 4a
7a cc c0 74 1d 09
"""
        frame_short = frame_short.replace(' ', '').replace('\n', '')
        self.frame_short = np.frombuffer(bytes.fromhex(frame_short),
                                         dtype='uint8')

        self.frames = (self.frame_long, self.frame_short)

        frame_long_out = """
a464829c8c4060a4a66060a6406f00f01642020001004200000000
000000000000000000000000000000000000000000000000000000
000000000000000000000000000000001b1bff671f20250eaab140
60f43c01002400f01c
"""
        frame_long_out = frame_long_out.replace('\n', '')
        self.frame_long_out = np.frombuffer(bytes.fromhex(frame_long_out),
                                            dtype='uint8')
        frame_short_out = """
a464829c8c4060a4a66060a6406f00f0e1ff020001000300002606
"""
        frame_short_out = frame_short_out.replace('\n', '')
        self.frame_short_out = np.frombuffer(bytes.fromhex(frame_short_out),
                                             dtype='uint8')
        self.frames_out = (self.frame_long_out, self.frame_short_out)

        self.sync = np.array([80, 114, 246,  75,  45, 144, 177, 245],
                             dtype='uint8')
        self.tb = gr.top_block()

    def tearDown(self):
        self.tb = None

    def test_usp_fec_deframer(self):
        """Inserts a long and a short frame in the USP deframer

        It also check that the output is correct"""
        frames = [np.concatenate([self.sync, f, np.zeros(1500, dtype='uint8')])
                  for f in self.frames]
        frames = [2 * np.unpackbits(f).astype('float32') - 1 for f in frames]
        frames = np.concatenate(frames)
        vector = blocks.vector_source_f(frames, False, 1, [])
        deframer = usp_deframer()
        dbg = blocks.message_debug()

        self.tb.connect(vector, deframer)
        self.tb.msg_connect((deframer, 'out'), (dbg, 'store'))
        self.tb.start()
        self.tb.wait()

        for j, frame in enumerate(self.frames_out):
            out = np.array(pmt.u8vector_elements(pmt.cdr(dbg.get_message(j))),
                           dtype='uint8')
            np.testing.assert_equal(out, frame, 'Output frame does not match')


if __name__ == '__main__':
    gr_unittest.run(qa_usp_deframer)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from satellites.satyaml import yamlfiles

def rest_transmitters(trans):
    print('  Transmitters:')
    print()
    for name, data in trans.items():
        print(f"    * **{name}** ({data['frequency']*1e-6:.3f} MHz): {data['modulation']} modulation with {data['framing']} framing")

def rest_satyaml(satyaml):
    print(f"**{satyaml['name']}**")
    if satyaml.get('alternative_names'):
        print('  Alternative names: ', end = '')
        print(', '.join(satyaml.get('alternative_names')))
        print()
    print(f"  NORAD ID: {satyaml['norad']}")
    print()
    rest_transmitters(satyaml['transmitters'])
    print()
        
def main():
    print('.. _Supported satellites:')
    print()
    print('Supported satellites')
    print('====================')
    print()
    print("""This is a list of all the satellites supported by gr-satellites.
The list is auto-generated by reading the SatYAML files and using the script ``docs/generate_supported_satellites.py``.
""")
    print()
    
    ymls = yamlfiles.load_all_yaml()
    for yml in sorted(ymls, key = lambda x: x['name'].casefold()):
        rest_satyaml(yml)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt

# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites import encode_rs, decode_rs


class qa_rs(gr_unittest.TestCase):
    def setUp(self):
        self.tb = gr.top_block()
        self.dbg = blocks.message_debug()

    def tearDown(self):
        self.tb.msg_connect((self.encode, 'out'), (self.decode, 'in'))
        self.tb.msg_connect((self.decode, 'out'), (self.dbg, 'store'))

        pdu = pmt.cons(pmt.PMT_NIL,
                       pmt.init_u8vector(len(self.data), self.data))
        self.encode.to_basic_block()._post(pmt.intern('in'), pdu)
        self.encode.to_basic_block()._post(
            pmt.intern('system'),
            pmt.cons(pmt.intern('done'), pmt.from_long(1)))

        self.tb.start()
        self.tb.wait()

        result = pmt.u8vector_elements(pmt.cdr(self.dbg.get_message(0)))
        np.testing.assert_equal(self.data, result,
                                'Decoded data does not match encoder input')

        self.tb = None

    def test_conventional(self):
        self.encode = encode_rs(False, 1)
        self.decode = decode_rs(False, 1)
        self.data = np.random.randint(0, 256, 223, dtype='uint8')

    def test_dual(self):
        self.encode = encode_rs(True, 1)
        self.decode = decode_rs(True, 1)
        self.data = np.random.randint(0, 256, 223, dtype='uint8')

    def test_shortened(self):
        self.encode = encode_rs(False, 1)
        self.decode = decode_rs(False, 1)
        self.data = np.random.randint(0, 256, 100, dtype='uint8')

    def test_interleave(self):
        interleave = 5
        self.encode = encode_rs(False, interleave)
        self.decode = decode_rs(False, interleave)
        self.data = np.random.randint(0, 256, 150 * interleave, dtype='uint8')

    def test_custom_rs(self):
        self.encode = encode_rs(8, 0x11d, 1, 1, 16, 1)
        self.decode = decode_rs(8, 0x11d, 1, 1, 16, 1)
        self.data = np.random.randint(0, 256, 255 - 16, dtype='uint8')


if __name__ == '__main__':
    gr_unittest.run(qa_rs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Daniel Estevez <daniel@destevez.net>.
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.

import numpy as np
import matplotlib.pyplot as plt
from gnuradio import gr, digital, blocks, channels, analog
from satellites.components.demodulators import bpsk_demodulator, fsk_demodulator

RAND_SEED = 42

samp_rate = 48e3
sps = 5

class LockInSim(gr.top_block):
    def __init__(self, lockin_block, time):
        gr.top_block.__init__(self)

        nbits = int(time * samp_rate / sps)
        self.source = blocks.vector_source_b(list(map(int, np.random.randint(0, 2, nbits))), False)
        
        self.lockin_block = lockin_block
        self.sink = blocks.null_sink(gr.sizeof_float)

        self.connect(self.source, self.lockin_block, self.sink)

class BPSK(gr.hier_block2):
    def __init__(self, ebn0):
        gr.hier_block2.__init__(self, 'BPSK',
            gr.io_signature(1, 1, gr.sizeof_char),
            gr.io_signature(1, 1, gr.sizeof_float))

        f_offset = 50
        packet_length = 0.2
        
        self.pack = blocks.pack_k_bits_bb(8)
        self.bpsk_constellation = digital.constellation_bpsk().base()
        self.modulator = digital.generic_mod(
            constellation = self.bpsk_constellation,
            differential = False,
            samples_per_symbol = sps,
            pre_diff_code = True,
            excess_bw = 0.35,
            verbose = False,
            log = False)

        self.packet_tx = analog.sig_source_f(samp_rate, analog.GR_SQR_WAVE, 1/packet_length,
                                            1, 0, 0)
        self.packet_tx_c = blocks.float_to_complex(1)
        self.tx = blocks.multiply_vcc(1)
        
        spb = sps
        self.channel = channels.channel_model(np.sqrt(spb)/10**(ebn0/20), f_offset/samp_rate, 1.0, [0,0,1], RAND_SEED, False)

        self.demod = bpsk_demodulator(samp_rate/sps, samp_rate, iq = True, dump_path = '/tmp/lockin_bpsk')

        self.connect(self, self.pack, self.modulator, self.tx, self.channel, self.demod, self)
        self.connect(self.packet_tx, self.packet_tx_c, (self.tx, 1))

class FSK(gr.hier_block2):
    def __init__(self, ebn0):
        gr.hier_block2.__init__(self, 'FSK',
            gr.io_signature(1, 1, gr.sizeof_char),
            gr.io_signature(1, 1, gr.sizeof_float))

        f_offset = 500
        packet_length = 0.2

        deviation = 5000
        bt = 1
        self.pack = blocks.pack_k_bits_bb(8)
        self.modulator = digital.gfsk_mod(
            samples_per_symbol = sps,
            sensitivity = 2*np.pi*deviation/samp_rate,
            bt = bt,
            verbose = False,
            log = False)

        self.packet_tx = analog.sig_source_f(samp_rate, analog.GR_SQR_WAVE, 1/packet_length,
                                            1, 0, 0)
        self.packet_tx_c = blocks.float_to_complex(1)
        self.tx = blocks.multiply_vcc(1)
        
        spb = sps
        self.channel = channels.channel_model(np.sqrt(spb)/10**(ebn0/20), f_offset/samp_rate, 1.0, [0,0,1], RAND_SEED, False)

        self.demod = fsk_demodulator(samp_rate/sps, samp_rate, deviation = deviation, iq = True, dump_path = '/tmp/lockin_fsk')

        self.connect(self, self.pack, self.modulator, self.tx, self.channel, self.demod, self)
        self.connect(self.packet_tx, self.packet_tx_c, (self.tx, 1))

if __name__ == '__main__':
    fg = LockInSim(BPSK(10), 1)
    fg.run()
    fg = LockInSim(FSK(15), 1)
    fg.run()



#!/usr/bin/env python3

# Copyright 2023 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import argparse
import datetime

import numpy as np
import scipy.constants
import skyfield.api
from skyfield.api import wgs84, EarthSatellite


DAY_S = 24 * 3600


ts = skyfield.api.load.timescale()


def parse_args():
    parser = argparse.ArgumentParser(
        description='Generates a Doppler file from a TLE')
    parser.add_argument(
        '--tle-file', required=True,
        help='Input TLE file')
    parser.add_argument(
        '--output-file', required=True,
        help='Ouptut text file')
    parser.add_argument(
        '--unix-timestamp', required=True, type=float,
        help='Initial UNIX timestamp')
    parser.add_argument(
        '--time-step', default=0.1, type=float,
        help='Time step (seconds) [default=%(default)r]')
    parser.add_argument(
        '--duration', default=15*60, type=float,
        help='Doppler file duration (seconds) [default=%(default)r]')
    parser.add_argument(
        '--f-carrier', required=True, type=float,
        help='Carrier frequency (Hz)')
    parser.add_argument(
        '--lat', required=True, type=float,
        help='Groundstation latitude (degrees)')
    parser.add_argument(
        '--lon', required=True, type=float,
        help='Groundstation longitude (degrees)')
    parser.add_argument(
        '--alt', default=0.0, type=float,
        help='Groundstation altitude (meters) [default=%(default)r]')
    return parser.parse_args()


def main():
    args = parse_args()
    with open(args.tle_file) as f:
        lines = f.readlines()
    if len(lines) == 3:
        # drop first line (contains the name of the satellite)
        lines = lines[1:]
    if len(lines) != 2:
        raise RuntimeError('TLE file must have either 2 or 3 lines')
    unix_epoch = datetime.datetime(1970, 1, 1, tzinfo=skyfield.api.utc)
    satellite = EarthSatellite(lines[0], lines[1], 'satellite', ts)
    t0 = unix_epoch + datetime.timedelta(seconds=args.unix_timestamp)
    t0 = ts.from_datetime(t0.replace(tzinfo=skyfield.api.utc))
    t = t0 + np.arange(0, (args.duration + args.time_step) / DAY_S,
                       args.time_step / DAY_S)
    t = ts.tai_jd([s.tai for s in t])
    groundstation = wgs84.latlon(
        args.lat, args.lon, args.alt)
    topocentric = (satellite - groundstation).at(t)
    range_rate = topocentric.frame_latlon_and_rates(
        groundstation)[5].km_per_s * 1e3
    doppler = - range_rate / scipy.constants.c * args.f_carrier
    with open(args.output_file, 'w') as output_file:
        for s, f in zip(t, doppler):
            s = (s.utc_datetime() - unix_epoch).total_seconds()
            print(f'{s}\t{f}', file=output_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import sys

from gnuradio import gr, blocks, gr_unittest
import numpy as np
import pmt


# bootstrap satellites module, even from build dir
try:
    import python as satellites
except ImportError:
    pass
else:
    import sys
    sys.modules['satellites'] = satellites

from satellites.components.datasinks import telemetry_parser


class qa_telemetry_parser(gr_unittest.TestCase):
    def setUp(self):
        self.tb = gr.top_block()

    def tearDown(self):
        self.tb = None

    def test_instance(self):
        """Tries to create a telemetry parser instance

        It uses several combinations of parameters
        """
        definition = 'by02'
        telemetry_parser(definition)
        telemetry_parser(definition, file=sys.stderr)
        telemetry_parser(definition, file='/dev/null')
        telemetry_parser(definition, options='')
        telemetry_parser(definition, options='--telemetry_output /dev/null')


if __name__ == '__main__':
    gr_unittest.run(qa_telemetry_parser)



#!/usr/bin/env python2
#
# Copyright 2012,2013 Free Software Foundation, Inc.
#
# This file is part of GNU Radio
#
# GNU Radio is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# GNU Radio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GNU Radio; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
#

import math
import numpy
from gnuradio import gr, digital
from gnuradio import analog
from gnuradio import blocks
from gnuradio import filter
from gnuradio.filter import firdes
from gnuradio import channels
from gnuradio import fec
import sys

try:
    from scipy.special import erfc
except ImportError:
    print "Error: could not import scipy (http://www.scipy.org/)"
    sys.exit(1)

try:
    import pylab
except ImportError:
    print "Error: could not import pylab (http://matplotlib.sourceforge.net/)"
    sys.exit(1)

from lilacsat1_ber_bpsk import *
from lilacsat1_ber_viterbi import *
    
# Best to choose powers of 10
N_BITS = 1e8
SKIP = 1000
RAND_SEED = 42

def berawgn(EbN0):
    """ Calculates theoretical bit error rate in AWGN (for BPSK and given Eb/N0) """
    return 0.5 * erfc(math.sqrt(10**(float(EbN0)/10)))

class BitErrors(gr.hier_block2):
    """test"""
    def __init__(self):
        gr.hier_block2.__init__(self, "BitErrors",
                gr.io_signature(1, 1, gr.sizeof_char),
                gr.io_signature(1, 1, gr.sizeof_int))

        intdump_decim = min(int(N_BITS/10), 100000)
        self.connect(self,
                     blocks.skiphead(gr.sizeof_char, SKIP),
                     blocks.not_bb(),
                     blocks.and_const_bb(1),
                     blocks.uchar_to_float(),
                     blocks.integrate_ff(intdump_decim),
                     blocks.multiply_const_ff(1.0/intdump_decim),
                     self)

class BERAWGNSimu(gr.top_block):
    " This contains the simulation flow graph "
    def __init__(self, EbN0, viterbi=False):
        gr.top_block.__init__(self)

        self.sps = 5
        alpha = 0.35

        const = digital.constellation_bpsk().base()
        modulator = digital.generic_mod(
          constellation=const,
          differential=False,
          samples_per_symbol=self.sps,
          pre_diff_code=True,
          excess_bw=alpha,
          verbose=False,
          log=False,
          )
        channel = channels.channel_model(
        	noise_voltage=self.EbN0_to_noise_voltage(EbN0, viterbi),
        	frequency_offset=0,
        	epsilon=1.0,
        	taps=(0, 0, (1+1j)/numpy.sqrt(2), ),
        	noise_seed=RAND_SEED,
        	block_tags=False
        )
        self.sink  = blocks.vector_sink_f()
        biterrors = BitErrors()

        dut = lilacsat1_ber_viterbi() if viterbi else lilacsat1_ber_bpsk()

        pack = blocks.pack_k_bits_bb(8)
        descrambler = digital.descrambler_bb(0x21, 0x00, 16)
        self.connect(blocks.vector_source_b([1], repeat=True),
                     blocks.head(gr.sizeof_char, int(N_BITS)),
                     digital.scrambler_bb(0x21, 0x00, 16),
                     digital.diff_encoder_bb(2),
                     pack)
        self.connect(modulator, channel,
                     blocks.multiply_const_cc(0.1), # we set some amplitude to test the agc # signal amplitude 1 seems very important
                     dut,
                     digital.diff_decoder_bb(2),
                     descrambler)
        self.connect(biterrors, self.sink)
        
        if viterbi:
            deinterleave_viterbi = blocks.deinterleave(gr.sizeof_char)
            interleave_viterbi = blocks.interleave(gr.sizeof_char)
            self.connect(pack,
                         fec.encode_ccsds_27_bb(),
                         deinterleave_viterbi)
            self.connect((deinterleave_viterbi, 0),
                         (interleave_viterbi, 1))
            self.connect((deinterleave_viterbi, 1),
                         blocks.not_bb(),
                         blocks.and_const_bb(1),
                         (interleave_viterbi, 0))
            self.connect(interleave_viterbi,
                         blocks.pack_k_bits_bb(8),
                         modulator)
            descrambler2 = digital.descrambler_bb(0x21, 0x00, 16)
            self.connect((dut, 1),
                        digital.diff_decoder_bb(2),
                        descrambler2)
            or2 = blocks.or_bb()
            self.connect(descrambler, or2)
            self.connect(descrambler2, (or2, 1))
            self.connect(or2, biterrors)
            #self.sinkviterbi1 = blocks.vector_sink_b()
            #self.sinkviterbi2 = blocks.vector_sink_b()
            #self.connect(descrambler, self.sinkviterbi1)
            #self.connect(descrambler2, self.sinkviterbi2)

        else:
            self.connect(pack, modulator)
            self.connect(descrambler, biterrors)            

    def EbN0_to_noise_voltage(self, EbN0, viterbi):
        """ Converts Eb/N0 to a complex noise voltage (assuming unit symbol power) """
        spb = float(self.sps)*2.0 if viterbi else float(self.sps) # samples per bit
        return 1.0 / math.sqrt(1/spb * 10**(float(EbN0)/10))

def simulate_ber(EbN0, viterbi = False):
    """ All the work's done here: create flow graph, run, read out BER """
    print "Eb/N0 = %d dB" % EbN0
    fg = BERAWGNSimu(EbN0, viterbi)
    fg.run()
    data = fg.sink.data()
    factor = 3.0 if viterbi else 6.0 # each bit error produces 6 errors, taking into account differential coding and scrambler
    return numpy.sum(data)/(factor*len(data))

if __name__ == "__main__":
    EbN0_min = 0
    EbN0_max = 12
    EbN0_range = range(EbN0_min, EbN0_max+1)
    ber_theory = [berawgn(x)      for x in EbN0_range]
    print "Simulating uncoded BPSK..."
    ber_simu_bpsk   = [simulate_ber(x, False) for x in EbN0_range]
    print "Simulating Viterbi..."
    ber_simu_viterbi   = [simulate_ber(x, True) for x in EbN0_range]

    f = pylab.figure()
    s = f.add_subplot(1,1,1)
    s.semilogy(EbN0_range, ber_theory, 'g-.', label="Theoretical uncoded BPSK")
    s.semilogy(EbN0_range, ber_simu_bpsk, 'b-o', label="Simulated uncoded BPSK")
    s.semilogy(EbN0_range, ber_simu_viterbi, 'r-o', label="Simulated Viterbi")
    s.set_title('BER Simulation')
    s.set_xlabel('Eb/N0 (dB)')
    s.set_ylabel('BER')
    s.legend()
    s.grid()
    pylab.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019-2023 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, filter, analog
from gnuradio.filter import firdes

from .fsk_demodulator import fsk_demodulator
from ...utils.options_block import options_block


class afsk_demodulator(gr.hier_block2, options_block):
    """
    Hierarchical block to demodulate AFSK.

    The input can be either IQ or real. For IQ input, it is assumed
    that the data is FM modulated, so FM demodulation is performed.
    For real input, it is assumed that the data is already FM
    demodulated.

    Args:
        baudrate: Baudrate in symbols per second (float)
        sample_rate: Sample rate in samples per second (float)
        iq: Whether the input is IQ or real (bool)
        af_carrier: Audio frequency carrier in Hz (float)
        deviation: Deviation in Hz, negative inverts the sidebands (float)
        dump_path: Path to dump internal signals to files (str)
        options: Options from argparse
    """
    def __init__(self, baudrate, samp_rate, iq, af_carrier,
                 deviation, dump_path=None, options=None,
                 fm_deviation=None):
        gr.hier_block2.__init__(
            self,
            'afsk_demodulator',
            gr.io_signature(1, 1,
                            gr.sizeof_gr_complex if iq else gr.sizeof_float),
            gr.io_signature(1, 1, gr.sizeof_float))
        options_block.__init__(self, options)

        # The FM deviation is used to apply Carson's rule in a low-pass filter
        # in the IQ case. In the FM demodulated case it is ignored.
        if fm_deviation is None:
            fm_deviation = self.options.fm_deviation

        if iq:
            # Note that deviation can be negative to encode that the
            # low tone corresponds to the symbol 1 and the high tone
            # corresponds to the symbol 0.
            carson_cutoff = fm_deviation + af_carrier + abs(deviation)
            self.demod = analog.quadrature_demod_cf(1)
            if carson_cutoff >= samp_rate / 2:
                # Sample rate is already narrower than Carson's
                # bandwidth. Do not filter
                self.connect(self, self.demod)
            else:
                # Sample rate is wider than Carson's bandwidth.
                # Lowpass filter before demod.
                fir_taps = firdes.low_pass(
                    1, samp_rate, carson_cutoff, 0.1 * carson_cutoff)
                self.demod_filter = filter.fir_filter_ccf(1, fir_taps)
                self.connect(self, self.demod_filter, self.demod)
        else:
            self.demod = self

        filter_cutoff = 2 * abs(deviation)
        filter_transition = 0.1 * abs(deviation)
        taps = firdes.low_pass(1, samp_rate, filter_cutoff, filter_transition)
        self.xlating = filter.freq_xlating_fir_filter_fcf(
            1, taps, af_carrier, samp_rate)

        self.fsk = fsk_demodulator(
            baudrate, samp_rate, deviation=deviation, iq=True,
            dc_block=False, dump_path=dump_path, options=options)

        self.connect(self.demod, self.xlating, self.fsk, self)

    _default_fm_deviation_hz = 3000

    @classmethod
    def add_options(cls, parser):
        """Adds AFSK demodulator options to the argparse parser"""
        fsk_demodulator.add_options(parser)
        parser.add_argument(
            '--fm-deviation', type=float,
            default=cls._default_fm_deviation_hz,
            help='FM deviation (Hz) [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital
import pmt

from ... import decode_ra_code
from ...crcs import crc16_arc, crc16_ccitt_false
from ...hier.sync_to_pdu_soft import sync_to_pdu_soft
from ...utils.options_block import options_block


_syncword = {
    'SMOG-P': '0010110111010100',
    'SMOG-1': '001011011101010001100011110001010011010110011001',
    'MRC-100': '11100011000111001001110110101110',
}


fec_sizes = {126: 252, 128: 260, 256: 514}


class mrc100_crc_reorder(gr.basic_block):
    """
    Puts the MRC-100 CRC at the end of the frame.

    The CRC is in bytes 1 and 2 of the frame, and should
    be computed over bytes 0, 3, 4, 5, .... This block rerranges
    the CRC to place it at the end of the frame.
    """
    def __init__(self, inverse=False):
        gr.basic_block.__init__(
            self,
            name='mrc100_crc_reorder',
            in_sig=[],
            out_sig=[])
        self.inverse = inverse
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        msg = pmt.u8vector_elements(msg)
        if self.inverse:
            msg = msg[:1] + msg[-2:] + msg[1:-2]
        else:
            msg = msg[:1] + msg[3:] + msg[1:3]
        msg = pmt.init_u8vector(len(msg), msg)
        msg_pmt = pmt.cons(pmt.car(msg_pmt), msg)
        self.message_port_pub(pmt.intern('out'), msg_pmt)


class smogp_ra_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the SMOG-P RA FEC frames

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        frame_size: size of the frame before FEC (int)
        variant: variant of the protocol to use ('SMOG-P', 'SMOG-1'
          or 'MRC-100') (str)
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, frame_size, variant='SMOG-P',
                 syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'smogp_ra_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.ra_syncword_threshold
        if syncword_threshold < 0:
            syncword_threshold = self._default_sync_threshold[variant]
        syncword = _syncword[variant]
        self.deframer = sync_to_pdu_soft(
            packlen=fec_sizes[frame_size] * 8, sync=syncword,
            threshold=syncword_threshold)
        self.fec = decode_ra_code(frame_size)
        if variant == 'SMOG-1':
            # CRC-16 ARC
            self.crc = crc16_arc(discard_crc=False)
        elif variant == 'MRC-100':
            self.crc_reorder = mrc100_crc_reorder()
            self.crc = crc16_ccitt_false(
                swap_endianness=True, discard_crc=False)
            self.crc_reorder2 = mrc100_crc_reorder(inverse=True)

        self.connect(self, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.fec, 'in'))
        if variant == 'SMOG-1':
            self.msg_connect((self.fec, 'out'), (self.crc, 'in'))
            self.msg_connect((self.crc, 'ok'), (self, 'out'))
        elif variant == 'MRC-100':
            self.msg_connect((self.fec, 'out'),
                             (self.crc_reorder, 'in'))
            self.msg_connect((self.crc_reorder, 'out'),
                             (self.crc, 'in'))
            self.msg_connect((self.crc, 'ok'),
                             (self.crc_reorder2, 'in'))
            self.msg_connect((self.crc_reorder2, 'out'),
                             (self, 'out'))
        else:
            self.msg_connect((self.fec, 'out'), (self, 'out'))

    _default_sync_threshold = {
        'SMOG-P': 0,
        'SMOG-1': 6,
        'MRC-100': 4,
    }

    @classmethod
    def add_options(cls, parser):
        """
        Adds SMOG-P RA deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--ra_syncword_threshold', type=int,
            default=-1,
            help='RA syncword bit errors [default=%(default)r]')
